#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <time.h>
#include <float.h>
#include "lr1_problem.h"
#include "misc.h"
#include "linear.h"

double evaluate_testing(double *wall, int n, const problem *prob_t)
{
	int l = prob_t->l;
	int *ty = prob_t->y;
	feature_node **tx = prob_t->x;

	int correct = 0;
	for(int i=0; i<l; i++)
	{
		int idx;
		double dec_value = 0;
		feature_node *txi = tx[i];
		while(txi->index!=-1 && txi->index<=n)
		{
			idx = txi->index;
			dec_value += wall[idx-1]*txi->value;
			txi++;
		}
		if(dec_value>0 && ty[i]==1)
			correct++;
		else if(dec_value<=0 && ty[i]==-1)
			correct++;
	}

	return (double)correct/l;
}

void transport(const problem *prob, double **x_val, int **x_ind, int **x_ptr)
{
	int n = prob->n;
	//int *y = prob->y;
	int *nl = new int[n];

	int i;
	for(i=0; i<n; i++)
		nl[i] = 0;
	int elements = 0;

	for(int i=0; i<prob->l; i++)
	{
		feature_node *s = prob->x[i];
		while(s->index != -1)
		{
			int ind = s->index-1;
			++nl[ind];
			++s;
			++elements;
		}
	}

	*x_val = Malloc(double, elements);
	*x_ind = Malloc(int, elements);
	*x_ptr = Malloc(int, n+1);

	int *t_ptr = Malloc(int, n+1);
	(*x_ptr)[0] = 0;	
	t_ptr[0] = 0;
	for(i=1; i<n+1 ;i++)
	{
		(*x_ptr)[i] = (*x_ptr)[i-1] + nl[i-1];
		t_ptr[i] = (*x_ptr)[i];
	}

	for(i=0; i<prob->l; i++)
	{
		feature_node *s = prob->x[i];
		while(s->index != -1)
		{
			int ind = s->index-1;
			(*x_val)[t_ptr[ind]] = s->value;
			(*x_ind)[t_ptr[ind]] = i;
			++t_ptr[ind];
			++s;
		}
	}
}

double lr1_problem::grad_condiction(double *g, double *w)
{
	double gc_norm2 = 0;
	double wall, gall;
	for(int i=0;i<n;i++){
		wall = w[i] - w[n+i];
		gall = g[i] -1;
		if(wall >0)
			gc_norm2 += (gall+1)*(gall+1);
		else if(wall < 0)
			gc_norm2 += (gall-1)*(gall-1);
		else if(fabs(gall) > 1) 
			gc_norm2 += (fabs(gall) -1)*(fabs(gall)-1);
	}
	return sqrt(gc_norm2);
}

lr1_problem::lr1_problem(const problem *prob, const problem *prob_t, double Cp, double Cn)
{
	int i;
	int *y = prob->y;

	this->l = prob->l;
	this->n = prob->n;
	this->prob = prob;
	this->prob_t = prob_t;

	C = Malloc(double, l);
	for(i=0; i<l; i++)
	{
		if (y[i] == 1)
			C[i] = Cp;
		else
			C[i] = Cn;
	}

	transport(prob, &x_val, &x_ir, &x_jc);
}

lr1_problem::~lr1_problem()
{
	free(x_val);
	free(x_jc);
	free(x_ir);
	free(C);
}


void lr1_problem::Xv(double *v, double *Xv)
{
	int i, j;
	//clock_t start, now;
	//start = clock();
	for(i=0;i<l;i++) Xv[i]=0;
	for(i=0;i<n;i++) 
		for(j=x_jc[i];j<x_jc[i+1];j++)
			Xv[x_ir[j]] += x_val[j]*v[i];
	//now = clock();
	//info("Xv time :%lf\n", (double)(now-start)/CLOCKS_PER_SEC);
}

void lr1_problem::XTv(double *v, double *XTv)
{
	int i, j;
	//clock_t start, now;
	//start = clock();
	for(i=0;i<n;i++)
	{
		XTv[i] = 0;
		for(j=x_jc[i];j<x_jc[i+1];j++)
			XTv[i] += v[x_ir[j]]*x_val[j];
	}
	//now = clock();
	//info("XTv time :%lf\n", (double)(now-start)/CLOCKS_PER_SEC);
}

// z cannot be changed after Xv
// but z is changed in grad
double lr1_problem::func(double *w, double *z)
{
	int i;
	int *y = prob->y;
	double f=0, yz;
	double* wall = Malloc(double, n);

	for(i=0; i<n; ++i)
	{
		wall[i] = w[i] - w[i+n];
		f = f + w[i] + w[i+n];
	}
	//f = f - w[n-1] - w[2*n-1]; //exclude bias term

	Xv(wall, z);

	for(i=0;i<l;i++)
	{
		yz = y[i]*z[i];
		if (yz >= 0)
			f += C[i]*log(1 + exp(-yz));
		else
			f += C[i]*(-yz+log(1 + exp(yz)));
	}

/*
#if DEBUG & 4
	info("func:\n");
	for(i=0; i<n; ++i)
		info("%lf ", wall[i]);
	info("\nfuncval: %lf\n", f);
#endif
*/

	free(wall);
	return(f);
	// later need better implementation to avoid numerical inaccuracy
}

void lr1_problem::grad(double *z, double *g, double *D)
{
	int i;
	int *y = prob->y;
	if(!z) fprintf(stderr, "error: z is not mallocated in grad()");

	for(i=0;i<l;i++)
	{
		z[i] = 1/(1 + exp(-y[i]*z[i]));
		D[i] = z[i]*(1-z[i]);
		z[i] = C[i]*(z[i]-1)*y[i];
	}
	XTv(z, g);	

	for(i=0;i<n;i++)
	{
		g[i+n] = 1 - g[i];
		g[i] += 1;
	}
	//g[2*n-1] = - g[n-1];
/*
#if DEBUG & 4
	info("grad:\n");
	for(i=0; i<2*n; ++i)
		info("%lf ", g[i]);
	info("\n");
#endif
*/
}

void lr1_problem::Hv(double *D, double *s, double *Hs)
{
	int i;
	double *wa = Malloc(double, l);
	double *sall = Malloc(double, n);

	for(i=0; i<n; ++i)
		sall[i] = s[i] - s[i+n];

	Xv(sall, wa);
	for(i=0;i<l;i++) 
		wa[i] = C[i]*D[i]*wa[i];

	XTv(wa, Hs);	
	for(i=0;i<n;i++) 
		Hs[i+n] = -Hs[i];

/*
#if DEBUG & 4
	info("Hv: s\n");
	for(i=0; i<2*n; ++i)
		info("%lf ", s[i]);
	info("\nHv: Hs\n");
	for(i=0; i<2*n; ++i)
		info("%lf ", Hs[i]);
	info("\n");
#endif
*/

	free(wa);
	free(sall);
}

void lr1_problem::subHv(int len, int *ind, double *D, double *s, double *Hs)
{
	int i, j;
	int pivot;
	int up, down;

	double *wa;
	wa = Malloc(double, l);
	for(i=0; i<l; ++i)
		wa[i] = 0;

	//clock_t start, now;
	/* possible speed up, but maybe worse @@
	int *upind, *downind;
	upind = Malloc(int, n); 
	downind = Malloc(int, n);

	for(i=0; i<n; ++i)
		upind[i] = downind[i] = -1;

	for(i=0; i<len; ++i){
		if(ind[i] < n)
			upind[ind[i]] = i;
		else
			downind[ind[i]-n] = i;
	}
	*/

	for(i=0; i<len; ++i)
		if(ind[i] >= n)
			break;
	pivot = i;

	//start = clock();
	//X(sup_f - sdown_f)
	up = 0;
	down = pivot;
	while(up < pivot && down < len)
	{
		if(ind[up] < ind[down]-n)
		{
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; ++j)
				wa[x_ir[j]] += x_val[j] * s[up];
			++up;
		}
		else if(ind[up] > ind[down]-n)
		{
			for(j=x_jc[ind[down]-n]; j<x_jc[ind[down]-n+1]; ++j)
				wa[x_ir[j]] -= x_val[j] * s[down];
			++down;
		}
		else
		{
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; ++j)
				wa[x_ir[j]] += x_val[j] * (s[up] - s[down]);
			++up;
			++down;
		}
	}
	if(up < pivot)
	{
		for(i=up; i<pivot; ++i)
			for(j=x_jc[ind[i]]; j<x_jc[ind[i]+1]; ++j)
				wa[x_ir[j]] += x_val[j] * s[i];
	}
	if(down < len)
	{
		for(i=down; i<len; ++i)
			for(j=x_jc[ind[i]-n]; j<x_jc[ind[i]-n+1]; ++j)
				wa[x_ir[j]] -= x_val[j] * s[i];
	}

	//now = clock();
	//info("subHv time :%lf\n", (double)(now-start)/CLOCKS_PER_SEC);

	//CDX(sup_f-sdown_f)
	for(i=0; i<l; ++i)
		wa[i] = wa[i] * C[i] * D[i];

	//start = clock();
	//CXTDX(sup_f-sdown_f)
	up = 0;
	down = pivot;
	while(up < pivot && down < len)
	{
		if(ind[up] < ind[down]-n)
		{
			Hs[up] = 0;
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; ++j)
				Hs[up] += x_val[j] * wa[x_ir[j]];
			++up;
		}
		else if(ind[up] > ind[down]-n)
		{
			Hs[down] = 0;
			for(j=x_jc[ind[down]-n]; j<x_jc[ind[down]-n+1]; ++j)
				Hs[down] -= x_val[j] * wa[x_ir[j]];
			++down;
		}
		else
		{
			Hs[up] = Hs[down] = 0;
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; ++j)
			{
				Hs[up] += x_val[j] * wa[x_ir[j]];
				Hs[down] -= x_val[j] * wa[x_ir[j]];
			}
			++up;
			++down;
		}
	}
	if(up < pivot)
	{
		for(i=up; i<pivot; ++i)
		{
			Hs[i] = 0;
			for(j=x_jc[ind[i]]; j<x_jc[ind[i]+1]; ++j)
				Hs[i] += x_val[j] * wa[x_ir[j]];
		}
	}
	if(down < len)
	{
		for(i=down; i<len; ++i)
		{
			Hs[i] = 0;
			for(j=x_jc[ind[i]-n]; j<x_jc[ind[i]-n+1]; ++j)
				Hs[i] -= x_val[j] * wa[x_ir[j]];
		}
	}
	//now = clock();
	//info("subHTv time :%lf\n", (double)(now-start)/CLOCKS_PER_SEC);

/*
#if DEBUG & 4
	info("subHv\n");
	info("len:%d\n", len);
	for(i=0; i<len; ++i)
		info("%d: %lf\n", ind[i], s[i]);
	info("wa\n");
	for(i=0; i<l; ++i)
		info("%lf\n", wa[i]);
	info("value:\n");
	for(i=0; i<len; ++i)
		info("%lf\n", Hs[i]);
#endif
*/

	/*
	free(upind);
	free(downind);
	*/
	free(wa);
}

double lr1_problem::testing(double *w)
{
	int i;
	double *wall = Malloc(double, n);

	double accuracy = 0;
	for(i=0; i<n; i++)
		wall[i] = w[i]-w[i+n];
	accuracy = evaluate_testing(wall, n, prob_t);

	free(wall);
	return accuracy;
}


//XXX
//for L1-Regularized L2-loss SVM
//
l2svm_problem::l2svm_problem(const problem *prob, const problem *prob_t, double Cp, double Cn)
{
	int i;
	int *y = prob->y;

	this->l = prob->l;
	this->n = prob->n;
	this->prob = prob;
	this->prob_t = prob_t;

	C = new double[l];
	I = new int[l];

	for(i=0; i<l; i++)
	{
		if (y[i] == 1)
			C[i] = Cp;
		else
			C[i] = Cn;
	}

	transport(prob, &x_val, &x_ir, &x_jc);
}

l2svm_problem::~l2svm_problem()
{
	delete[] C;
	delete[] I;
	free(x_val);
	free(x_jc);
	free(x_ir);
}


//XXX access by column format
void l2svm_problem::Xv(double *v, double *Xv)
{
	int i, j;
	for(i=0;i<l;i++) Xv[i]=0.0;
	for(i=0;i<n;i++) 
	{
		for(j=x_jc[i];j<x_jc[i+1];j++)
		{
			Xv[x_ir[j]] += x_val[j]*v[i];
		}
	}
	/*
	int i;
	feature_node **x = prob->x;

	for(i=0; i<l; i++)
	{
		feature_node *s = x[i];
		Xv[i] = 0.0;
		while(s->index != -1)
		{
			Xv[i] += v[s->index-1]*s->value;
			s++;
		}
	}
	*/
}

//XXX access by row format
void l2svm_problem::subXv(double *v, double *Xv)
{
	int i;
	feature_node **x = prob->x;

	for(i=0; i<sizeI; i++)
	{
		feature_node *s = x[I[i]];
		Xv[i] = 0.0;
		while(s->index != -1)
		{
			Xv[i] += v[s->index-1]*s->value;
			s++;
		}
	}
}

/*
void l2svm_problem::XTv(double *v, double *XTv)
{
	int i, j;
	for(i=0;i<n;i++)
	{
		XTv[i] = 0.0;
		for(j=x_jc[i];j<x_jc[i+1];j++){
			XTv[i] += v[x_ir[j]]*x_val[j];
		}
	}
	/\*
	int i;
	feature_node **x = prob->x;

	for(i=0; i<n; i++)
		XTv[i] = 0.0;
	for(i=0; i<l; i++)
	{
		feature_node *s = x[i];
		while(s->index != -1)
		{
			XTv[s->index-1] += v[i]*s->value;
			s++;
		}
	}
	*\/
}
*/

void l2svm_problem::subXTv(double *v, double *XTv)
{
	int i;
	int n = prob->n;
	feature_node **x = prob->x;

	for(i=0; i<n; i++)
		XTv[i] = 0.0;
	for(i=0; i<sizeI; i++)
	{
		feature_node *s = x[I[i]];
		while(s->index != -1)
		{
			XTv[s->index-1] += v[i]*s->value;
			s++;
		}
	}
}


// [z] cannot be changed after Xv
// but [z] will be changed in grad()
double l2svm_problem::func(double *w, double *z)
{
	int i;
	double f = 0.0;
	int *y = prob->y;
	double *wall = Malloc(double, n);

	for(i=0; i<n; i++)
	{
		wall[i] = w[i] - w[i+n];
		f = f + w[i] + w[i+n];
	}
	//f = f - w[n-1] - w[2*n-1]; //exclude bias term

	Xv(wall, z);
	for(i=0; i<l; i++)
	{
		z[i] = y[i]*z[i];
		double d = z[i] - 1;
		if (d < 0)
			f += C[i]*d*d;

		//XXX
		/*
		double yz = y[i]*z[i];
		if (yz >= 0.0)
			f += C[i]*log(1 + exp(-yz));
		else
			f += C[i]*(-yz+log(1 + exp(yz)));
		*/
	}

	free(wall);
	return(f);
}

double l2svm_problem::grad_condiction(double *g, double *w)
{
	double gc_norm2 = 0;
	double wall, gall;
	for(int i=0;i<n;i++){
		wall = w[i] - w[n+i];
		gall = g[i] -1;
		if(wall >0)
			gc_norm2 += (gall+1)*(gall+1);
		else if(wall < 0)
			gc_norm2 += (gall-1)*(gall-1);
		else if(fabs(gall) > 1) 
			gc_norm2 += (fabs(gall) -1)*(fabs(gall)-1);
	}
	return sqrt(gc_norm2);
}

void l2svm_problem::grad(double *z, double *g, double *D)
{
	if(!z) fprintf(stderr, "error: z is not mallocated in grad()");

	int i;
	int *y = prob->y;

	sizeI = 0;
	for(i=0; i<l; i++)
	{
		if (z[i] < 1)
		{
			z[sizeI] = C[i]*y[i]*(z[i]-1);
			I[sizeI] = i;
			sizeI ++;
		}
		//XXX
		/*
		z[i] = 1/(1 + exp(-y[i]*z[i]));
		D[i] = z[i]*(1-z[i]);
		z[i] = C[i]*(z[i]-1)*y[i];
		*/
	}
	subXTv(z, g);	

	for(i=0; i<n; i++)
	{
		g[i+n] = 1 - 2*g[i];
		g[i] = 1 + 2*g[i];
	}
	//g[2*n-1] = -g[n-1];
}

void l2svm_problem::Hv(double *D, double *s, double *Hs)
{
	int i;
	double *sall = Malloc(double, n);
	double *wa = Malloc(double, l);

	for(i=0; i<n; i++)
		sall[i] = s[i]-s[i+n];

	subXv(sall, wa);
	for(i=0; i<sizeI; i++) 
		wa[i] = C[I[i]]*wa[i];

	subXTv(wa, Hs);
	for(i=0; i<n; i++) 
	{
		Hs[i] = 2*Hs[i];
		Hs[i+n] = -Hs[i];
	}

	free(sall);
	free(wa);
}

void l2svm_problem::subHv(int len, int *ind, double *D, double *s, double *Hs)
{
	//XXX temporary
	int i;
	double *sAll = Malloc(double, 2*n);
	double *HsAll = Malloc(double, 2*n);
	for(i=0; i<n; i++)
	{
		sAll[i] = 0.0;
		sAll[i+n] = 0.0;
	}
	for(i=0; i<len; i++)
	{
		sAll[ind[i]] = s[i];
	}
	Hv(D, sAll, HsAll);
	for(i=0; i<len; i++)
	{
		Hs[i] = HsAll[ind[i]];
	}
	free(sAll);
	free(HsAll);

	/*
	int i, j, k;
	int pivot;
	int up, down;
	double *wa = Malloc(double, l);

	for(i=0; i<l; i++)
		wa[i] = DBL_MAX;
	for(k=0; k<sizeI; k++)
		wa[I[k]] = 0.0;

	for(i=0; i<len; i++)
	{
		if(ind[i] >= n)
			break;
	}
	pivot = i;

	up = 0;
	down = pivot;
	while(up<pivot && down<len)
	{
		if(ind[up] < ind[down]-n)
		{
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					wa[x_ir[j]] += x_val[j]*s[up];
			}
			++up;
		}
		else if(ind[up] > ind[down]-n)
		{
			for(j=x_jc[ind[down]-n]; j<x_jc[ind[down]-n+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					wa[x_ir[j]] -= x_val[j]*s[down];
			}
			++down;
		}
		else
		{
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					wa[x_ir[j]] += x_val[j]*(s[up]-s[down]);
			}	
			++up;
			++down;
		}
	}
	if(up < pivot)
	{
		for(i=up; i<pivot; i++)
		{
			for(j=x_jc[ind[i]]; j<x_jc[ind[i]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					wa[x_ir[j]] += x_val[j]*s[i];
			}
		}
	}
	if(down < len)
	{
		for(i=down; i<len; i++)
		{
			for(j=x_jc[ind[i]-n]; j<x_jc[ind[i]-n+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					wa[x_ir[j]] -= x_val[j]*s[i];
			}
		}
	}


	//
	for(i=0; i<sizeI; i++)
		wa[I[i]] = C[I[i]]*wa[I[i]];


	up = 0;
	down = pivot;
	while(up<pivot && down<len)
	{
		if(ind[up] < ind[down]-n)
		{
			Hs[up] = 0;
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					Hs[up] += x_val[j]*wa[x_ir[j]];
			}
			++up;
		}
		else if(ind[up] > ind[down]-n)
		{
			Hs[down] = 0;
			for(j=x_jc[ind[down]-n]; j<x_jc[ind[down]-n+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					Hs[down] -= x_val[j]*wa[x_ir[j]];
			}
			++down;
		}
		else
		{
			Hs[up] = Hs[down] = 0;
			for(j=x_jc[ind[up]]; j<x_jc[ind[up]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
				{
					Hs[up] += x_val[j]*wa[x_ir[j]];
					Hs[down] -= x_val[j]*wa[x_ir[j]];
				}
			}
			++up;
			++down;
		}
	}
	if(up < pivot)
	{
		for(i=up; i<pivot; i++)
		{
			Hs[i] = 0;
			for(j=x_jc[ind[i]]; j<x_jc[ind[i]+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					Hs[i] += x_val[j]*wa[x_ir[j]];
			}
		}
	}
	if(down < len)
	{
		for(i=down; i<len; i++)
		{
			Hs[i] = 0;
			for(j=x_jc[ind[i]-n]; j<x_jc[ind[i]-n+1]; j++)
			{
				if(wa[x_ir[j]] < DBL_MAX)
					Hs[i] -= x_val[j]*wa[x_ir[j]];
			}
		}
	}

	free(wa);
	*/
}


double l2svm_problem::testing(double *w)
{
	int i;
	double *wall = Malloc(double, n);
	double accuracy = 0.0;

	for(i=0; i<n; i++)
		wall[i] = w[i]-w[i+n];
	accuracy = evaluate_testing(wall, n, prob_t);

	free(wall);
	return accuracy;
}


